%{ 
  open Ast 
%}

%token MATRIX STRING CHAR LBRACK RBRACK
%token FLOAT /* STRING */
%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA
%token PLUS MINUS TIMES DIVIDE ASSIGN NOT
%token EQ NEQ LT LEQ GT GEQ TRUE FALSE AND OR
%token RETURN IF ELSE FOR WHILE INT BOOL VOID
%token <int> LITERAL
%token <bool> BLIT 
%token <string> ID FLOATLIT
%token <string> STRINGLIT
%token <char> CHARLIT
%token EOF

%start program
%type <Ast.program> program

%nonassoc NOELSE
%nonassoc ELSE
%right ASSIGN
%left OR
%left AND
%left EQ NEQ
%left LT GT LEQ GEQ
%left PLUS MINUS
%left TIMES DIVIDE
%right NOT

%%


program:
  decls EOF { $1 }

decls:
    /* nothing */ { [], [] }
 | decls vdecl { (($2 :: fst $1), snd $1) }
 | decls fdecl { (fst $1, ($2 :: snd $1)) }

fdecl:
   typ ID LPAREN formals_opt RPAREN LBRACE vdecl_list stmt_list RBRACE
     { { typ = $1;
         fname = $2;
         formals = List.rev $4;
         locals = List.rev $7;
         body = List.rev $8 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { $1 }

formal_list:
    typ ID                   { [($1,$2, Noexpr )]     }
  | formal_list COMMA typ ID { ($3,$4, Noexpr ) :: $1 }

typ:
    INT    { Int   }
  | BOOL   { Bool  }
  | FLOAT  { Float }
  | STRING { String }
  | CHAR   { Char }
  | MATRIX { Matrix }
  | VOID   { Void  }

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
   typ ID SEMI { ($1, $2, Noexpr) }
  | typ ID ASSIGN expr SEMI { ($1, $2, Asn($2,$4))}

stmt_list:
    /* nothing */  { [] }
  | stmt_list stmt { $2 :: $1 }

stmt:
    expr SEMI                               { Expr ($1)               }
  | RETURN expr_opt SEMI                    { Return ($2)             }
  | LBRACE stmt_list RBRACE                 { Block(List.rev $2)    }
  | IF LPAREN expr RPAREN stmt %prec NOELSE { If($3, $5, Block([])) }
  | IF LPAREN expr RPAREN stmt ELSE stmt    { If($3, $5, $7)        }
  | FOR LPAREN expr_opt SEMI expr SEMI expr_opt RPAREN stmt
                                            { For($3, $5, $7, $9)   }
  | WHILE LPAREN expr RPAREN stmt           { While($3, $5)         }

expr_opt:
    /* nothing */ { Noexpr }
  | expr          { $1 }

expr:
    LITERAL          { Literal($1)            }
  | FLOATLIT         { Fliteral($1)           }
  | BLIT             { BoolLit($1)            }
  | CHARLIT          { CharLit($1)          } 
  | STRINGLIT        { StringLit($1)          }
  | TRUE             { BoolLit(true)          }
  | FALSE            { BoolLit(false)         }
  | ID               { Id($1)                 }
  /* | ID matrx_addr              { GetMatrixVal($1,$2)    } */
  /* | ID matrx_addr ASSIGN expr  { SetMatrixVal($1,$2,$4) } */
  | LBRACK matrx_value RBRACK       { MatrixLit($2)          }
  | expr PLUS   expr { Binop($1, Add,   $3)   }
  | expr MINUS  expr { Binop($1, Sub,   $3)   }
  | expr TIMES  expr { Binop($1, Mult,  $3)   }
  | expr DIVIDE expr { Binop($1, Div,   $3)   }
  | expr EQ     expr { Binop($1, Equal, $3)   }
  | expr NEQ    expr { Binop($1, Neq,   $3)   }
  | expr LT     expr { Binop($1, Less,  $3)   }
  | expr LEQ    expr { Binop($1, Leq,   $3)   }
  | expr GT     expr { Binop($1, Greater, $3) }
  | expr GEQ    expr { Binop($1, Geq,   $3)   }
  | expr AND    expr { Binop($1, And,   $3)   }
  | expr OR     expr { Binop($1, Or,    $3)   }
  | MINUS expr %prec NOT { Unop(Neg, $2)      }
  | NOT expr         { Unop(Not, $2)          }
  | ID ASSIGN expr   { Asn($1, $3)            } /* handles assignment */
  | ID LPAREN seq_opt RPAREN { Call($1, $3)  }
  | LPAREN expr RPAREN { $2                   }

seq_opt:
    /* nothing */ { [] }
  | seq  { List.rev $1 }

seq:
  expr                                { [$1] }
|  seq COMMA expr                     { $3 :: $1 }

matrx_value:
  LBRACK seq RBRACK                   { [MatrixLit(List.rev $2)] }
| LBRACK seq RBRACK COMMA matrx_value { MatrixLit(List.rev $2)::$5 }
| LBRACK seq RBRACK matrx_value       { MatrixLit(List.rev $2)::$4 }